/*
 * Copyright (c) 2021, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 */

package org.wso2.cloud.secrets.utils;

import com.azure.identity.ClientSecretCredential;
import com.azure.identity.ClientSecretCredentialBuilder;
import com.azure.security.keyvault.certificates.CertificateClient;
import com.azure.security.keyvault.certificates.CertificateClientBuilder;
import com.azure.security.keyvault.certificates.models.KeyVaultCertificateWithPolicy;
import com.azure.security.keyvault.secrets.SecretClient;
import com.azure.security.keyvault.secrets.SecretClientBuilder;
import com.azure.security.keyvault.secrets.models.KeyVaultSecret;
import org.wso2.cloud.secrets.exception.EncryptionException;
import org.wso2.cloud.secrets.model.InputSecretIdentifiers;
import org.wso2.cloud.secrets.model.Secret;
import sun.security.provider.X509Factory;

import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.Base64;
import java.util.List;

/**
 * Utilities required for dealing with Azure Key Vault.
 *
 * @since 1.0.0
 */
public class AzureKeyVaultUtils {

    /**
     * Retrieve the defined plain text passwords from the source Azure Key Vault.
     * The Azure Key Vault Secrets to be retrieved are defined in the input secrets file.
     *
     * @param inputSecretIdentifiers input secrets which are to be encrypted and pushed to the target Key Vault
     * @param secretClient           a {@link SecretClient} instance for source Azure Key Vault access
     * @return List of input secrets which are to be encrypted and pushed to the target Key Vault
     */
    public static List<Secret> retrievePlainTextSecrets(
            InputSecretIdentifiers inputSecretIdentifiers, SecretClient secretClient) {

        if (inputSecretIdentifiers == null) {
            throw new EncryptionException("No input secret data");
        }

        List<Secret> secrets = new ArrayList<>();
        inputSecretIdentifiers.getSecrets().forEach(inputSecretIdentifier -> {
            KeyVaultSecret secret = secretClient.getSecret(
                    inputSecretIdentifier.getIdentifier(), inputSecretIdentifier.getVersion());
            secrets.add(new Secret(inputSecretIdentifier.getIdentifier(), secret.getValue()));
        });

        return secrets;
    }

    /**
     * Push the provided encrypted secrets to the target Azure Key Vault.
     *
     * @param secrets      encrypted secrets to be pushed to the target Key Vault
     * @param secretClient a {@link SecretClient} instance for target Azure Key Vault access
     */
    public static void pushSecretsToAzureKeyVault(List<Secret> secrets, SecretClient secretClient) {

        if (secrets == null) {
            throw new EncryptionException("No secrets to push to Azure Key Vault. List of secrets cannot be null.");
        }

        secrets.forEach(secret ->
                secretClient.setSecret(new KeyVaultSecret(secret.getIdentifier(), secret.getEncryptedPassword()))
        );
    }

    /**
     * Retrieve the public key from the Azure Key Vault Certificate resource.
     *
     * @param certificateClient a {@link CertificateClient} instance for target Azure Key Vault access
     * @return complete public key string
     */
    public static String getPublicKeyForEncryption(CertificateClient certificateClient) {

        String certificateName = System.getProperty(Constants.ENCRYPTION_PUBLIC_KEY_SYS_PROPERTY_KEY);
        KeyVaultCertificateWithPolicy certificate = certificateClient.getCertificate(certificateName);

        String certificateString = new String(Base64.getEncoder().encode(certificate.getCer()), StandardCharsets.UTF_8);

        return X509Factory.BEGIN_CERT + "\n" + certificateString + "\n" + X509Factory.END_CERT;
    }

    /**
     * Create {@link SecretClient} instance using the provided Azure Key Vault name.
     *
     * @param azureKeyVaultName name of the Azure Key Vault
     * @return created {@link SecretClient} instance using the provided Azure Key Vault name
     */
    public static SecretClient createSecretClient(String azureKeyVaultName) {

        return new SecretClientBuilder()
                .vaultUrl(Constants.HTTPS_PROTOCOL_PREFIX + azureKeyVaultName + Constants.AZURE_KEY_VAULT_DOMAIN)
                .credential(createClientSecretCredential()).buildClient();
    }

    /**
     * Create {@link CertificateClient} instance using the provided Azure Key Vault name.
     *
     * @param azureKeyVaultName name of the Azure Key Vault
     * @return created {@link CertificateClient} instance using the provided Azure Key Vault name
     */
    public static CertificateClient createCertificateClient(String azureKeyVaultName) {

        return new CertificateClientBuilder()
                .vaultUrl(Constants.HTTPS_PROTOCOL_PREFIX + azureKeyVaultName + Constants.AZURE_KEY_VAULT_DOMAIN)
                .credential(createClientSecretCredential()).buildClient();
    }

    /**
     * Create Client Secret Credential {@link ClientSecretCredential} instance for authentication using the
     * Azure Active Directory Service Principal details.
     *
     * @return generated Client Secret Credential {@link ClientSecretCredential} instance
     */
    private static ClientSecretCredential createClientSecretCredential() {

        String servicePrincipalID = System.getProperty(Constants.SERVICE_PRINCIPAL_FOR_APP_ID_SYS_PROPERTY_KEY);
        String servicePrincipalKey = System.getProperty(Constants.SERVICE_PRINCIPAL_FOR_APP_KEY_SYS_PROPERTY_KEY);
        String tenantID = System.getProperty(Constants.KEY_VAULT_TENANT_ID_SYS_PROPERTY_KEY);

        return new ClientSecretCredentialBuilder()
                .clientId(servicePrincipalID)
                .clientSecret(servicePrincipalKey)
                .tenantId(tenantID)
                .build();
    }
}
