/*
 * Copyright (c) 2021, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 */

package org.wso2.cloud.secrets.utils;

import com.google.gson.Gson;
import org.bouncycastle.jce.provider.BouncyCastleProvider;
import org.wso2.cloud.secrets.exception.EncryptionException;
import org.wso2.cloud.secrets.model.InputSecretIdentifiers;
import org.wso2.cloud.secrets.model.Secret;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.Reader;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;
import java.security.Security;
import java.security.cert.Certificate;
import java.security.cert.CertificateException;
import java.security.cert.CertificateFactory;
import java.util.List;
import javax.crypto.BadPaddingException;
import javax.crypto.Cipher;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.NoSuchPaddingException;
import javax.xml.bind.DatatypeConverter;

/**
 * Utilities required for encryption of provided plain text secrets.
 *
 * @since 1.0.0
 */
public class EncryptionUtils {

    /**
     * Encrypt the plain text password of each {@link Secret} of a list of {@link Secret} instances.
     * Uses the public key string for encryption.
     *
     * @param secrets           list of input {@link Secret} instances
     * @param certificateString certificate string of the public key used to encrypt the input plain text passwords
     */
    public static void encryptPlainTextSecretsUsingCertString(List<Secret> secrets, String certificateString) {

        if (secrets == null) {
            throw new EncryptionException("No input secret data to input");
        }

        if (certificateString == null || certificateString.isEmpty()) {
            throw new EncryptionException("No certificate provided");
        }

        Security.addProvider(new BouncyCastleProvider());

        //  check if the cryptographic algorithm to be used is defined via a system property
        //  if not set, use the default transformation algorithm (RSA)
        String transformation = System.getProperty(Constants.CIPHER_TRANSFORMATION_SYSTEM_PROPERTY);
        if (transformation == null) {
            transformation = "RSA";
        }

        //  initialize the Cipher used for the secret encryption process, using the public key string
        Cipher cipherForEncryption;
        InputStream stream = new ByteArrayInputStream(certificateString.getBytes(StandardCharsets.UTF_8));
        try {
            CertificateFactory certificateFactory = CertificateFactory.getInstance("X.509");
            Certificate certificate = certificateFactory.generateCertificate(stream);
            cipherForEncryption = Cipher.getInstance(transformation);
            cipherForEncryption.init(Cipher.ENCRYPT_MODE, certificate);
        } catch (CertificateException | NoSuchPaddingException
                | NoSuchAlgorithmException | InvalidKeyException exception) {
            throw new EncryptionException("Error when initializing the cipher", exception);
        }

        secrets.forEach(secret ->
                secret.setEncryptedPassword(
                        EncryptionUtils.performEncryption(cipherForEncryption, secret.getPassword()))
        );
    }

    /**
     * Encrypt the provided plain text password using the {@link Cipher}
     *
     * @param cipher            {@link Cipher} to be used for encryption
     * @param plainTextPassword the plain text password to be encrypted
     * @return encrypted password
     */
    private static String performEncryption(Cipher cipher, String plainTextPassword) {

        if ((plainTextPassword == null) || (plainTextPassword.isEmpty())) {
            throw new EncryptionException("Password is either null or an empty string");
        }

        String encodedValue;
        try {
            byte[] encryptedPassword = cipher.doFinal(plainTextPassword.getBytes(StandardCharsets.UTF_8));
            encodedValue = DatatypeConverter.printBase64Binary(encryptedPassword);
        } catch (BadPaddingException | IllegalBlockSizeException exception) {
            throw new EncryptionException("Error when encrypting the provided password", exception);
        }

        return encodedValue;
    }

    /**
     * Load the input secret identifier data from the source JSON file.
     *
     * @return List of identifiers of the input secret which are to be encrypted and pushed to the target Key Vault
     */
    public static InputSecretIdentifiers loadInputSecrets() {

        String secretsSourceFile = System.getProperty(Constants.INPUT_SECRETS_SOURCE_FILE_SYS_PROPERTY_KEY);
        Gson gson = new Gson();

        Reader reader;
        try {
            reader = Files.newBufferedReader(Paths.get(secretsSourceFile));
        } catch (IOException exception) {
            throw new EncryptionException("Unable to read the source file: " + secretsSourceFile, exception);
        }

        return gson.fromJson(reader, InputSecretIdentifiers.class);
    }

    /**
     * Validate the required input Java System Property values.
     */
    public static void validateSystemPropertyBasedInput() {

        validateSystemPropertyValue(Constants.INPUT_SECRETS_SOURCE_FILE_SYS_PROPERTY_KEY,
                "System property for source file with secret identifiers '" +
                        Constants.INPUT_SECRETS_SOURCE_FILE_SYS_PROPERTY_KEY + "' not defined or " +
                        "file name resolves to an empty string");
        validateSystemPropertyValue(Constants.SOURCE_AZURE_KEY_VAULT_NAME_SYS_PROPERTY_KEY,
                "System property for source Key Vault '" +
                        Constants.SOURCE_AZURE_KEY_VAULT_NAME_SYS_PROPERTY_KEY + "' not defined or " +
                        "source Key Vault name resolves to an empty string");
        validateSystemPropertyValue(Constants.TARGET_AZURE_KEY_VAULT_NAME_SYS_PROPERTY_KEY,
                "System property for target Key Vault '" +
                        Constants.TARGET_AZURE_KEY_VAULT_NAME_SYS_PROPERTY_KEY + "' not defined or " +
                        "target Key Vault name resolves to an empty string");
        validateSystemPropertyValue(Constants.ENCRYPTION_PUBLIC_KEY_SYS_PROPERTY_KEY,
                "System property for Azure Key Vault Certificate name '" +
                        Constants.ENCRYPTION_PUBLIC_KEY_SYS_PROPERTY_KEY + "' not defined or " +
                        "Certificate name resolves to an empty string");
        validateSystemPropertyValue(Constants.SERVICE_PRINCIPAL_FOR_APP_ID_SYS_PROPERTY_KEY,
                "System property for Azure AD Service Principal Application ID '" +
                        Constants.SERVICE_PRINCIPAL_FOR_APP_ID_SYS_PROPERTY_KEY + "' not defined or " +
                        "Service Principal ID resolves to an empty string");
        validateSystemPropertyValue(Constants.SERVICE_PRINCIPAL_FOR_APP_KEY_SYS_PROPERTY_KEY,
                "System property for Azure AD Service Principal Key '" +
                        Constants.SERVICE_PRINCIPAL_FOR_APP_KEY_SYS_PROPERTY_KEY + "' not defined or " +
                        "Service Principal Key resolves to an empty string");
        validateSystemPropertyValue(Constants.KEY_VAULT_TENANT_ID_SYS_PROPERTY_KEY,
                "System property for Azure Tenant ID '" +
                        Constants.KEY_VAULT_TENANT_ID_SYS_PROPERTY_KEY + "' not defined or " +
                        "Azure Key Vault Tenant ID resolves to an empty string");
    }

    /**
     * Validate a given Java System Property value is empty or null.
     *
     * @param propertyName name of the Java System Property
     * @param errorMessage error message to add if the value is empty or null
     */
    private static void validateSystemPropertyValue(String propertyName, String errorMessage) {

        if ((propertyName == null) || (propertyName.isEmpty())) {
            throw new EncryptionException("Invalid system property name: Name cannot be null or an empty string");
        }

        String propertyValue = System.getProperty(propertyName);
        if (propertyValue == null || propertyValue.isEmpty()) {
            if (errorMessage == null || errorMessage.isEmpty()) {
                throw new EncryptionException("System property " + propertyName + " not set");
            } else {
                throw new EncryptionException(errorMessage);
            }
        }
    }
}
